@@ -15,7 +15,7 @@
 import pickle, os, uuid
 from .mesonlib import MesonException, default_libdir, default_libexecdir, default_prefix
 
-version = '0.35.0'
+version = '0.37.1'
 backendlist = ['ninja', 'vs2010', 'vs2015', 'xcode']
 
 class UserOption:
@@ -36,11 +36,6 @@ class UserStringOption(UserOption):
     def validate(self, value):
         if not isinstance(value, str):
             raise MesonException('Value "%s" for string option "%s" is not a string.' % (str(value), self.name))
-        if self.name == 'prefix' and not os.path.isabs(value):
-            raise MesonException('Prefix option value \'{0}\' must be an absolute path.'.format(value))
-        if self.name in ('libdir', 'bindir', 'includedir', 'datadir', 'mandir', 'localedir') \
-            and os.path.isabs(value):
-            raise MesonException('Option %s must not be an absolute path.' % self.name)
 
     def set_value(self, newvalue):
         self.validate(newvalue)
@@ -154,18 +149,22 @@ class CoreData():
             raise RuntimeError('Tried to set unknown builtin option %s.' % optname)
 
 def load(filename):
-    with open(filename, 'rb') as f:
-        obj = pickle.load(f)
+    load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
+    try:
+        with open(filename, 'rb') as f:
+            obj = pickle.load(f)
+    except pickle.UnpicklingError:
+        raise MesonException(load_fail_msg)
     if not isinstance(obj, CoreData):
-        raise RuntimeError('Core data file is corrupted.')
+        raise MesonException(load_fail_msg)
     if obj.version != version:
-        raise RuntimeError('Build tree has been generated with Meson version %s, which is incompatible with current version %s.'%
+        raise MesonException('Build directory has been generated with Meson version %s, which is incompatible with current version %s.\nPlease delete this build directory AND create a new one.'%
                            (obj.version, version))
     return obj
 
 def save(obj, filename):
     if obj.version != version:
-        raise RuntimeError('Fatal version mismatch corruption.')
+        raise MesonException('Fatal version mismatch corruption.')
     with open(filename, 'wb') as f:
         pickle.dump(obj, f)
 
@@ -209,25 +208,35 @@ builtin_options = {
         'libdir'            : [ UserStringOption, 'Library directory.', default_libdir() ],
         'libexecdir'        : [ UserStringOption, 'Library executable directory.', default_libexecdir() ],
         'bindir'            : [ UserStringOption, 'Executable directory.', 'bin' ],
+        'sbindir'           : [ UserStringOption, 'System executable directory.', 'sbin' ],
         'includedir'        : [ UserStringOption, 'Header file directory.', 'include' ],
         'datadir'           : [ UserStringOption, 'Data file directory.', 'share' ],
         'mandir'            : [ UserStringOption, 'Manual page directory.', 'share/man' ],
+        'infodir'           : [ UserStringOption, 'Info page directory.', 'share/info' ],
         'localedir'         : [ UserStringOption, 'Locale data directory.', 'share/locale' ],
-    # Sysconfdir is a bit special. It defaults to ${prefix}/etc but nobody
-    # uses that. Instead they always set it manually to /etc. This default
-    # value is thus pointless and not really used but we set it to this
-    # for consistency with other systems.
+    # sysconfdir, localstatedir and sharedstatedir are a bit special. These defaults to ${prefix}/etc,
+    # ${prefix}/var and ${prefix}/com but nobody uses that. Instead they always set it
+    # manually to /etc, /var and /var/lib. This default values is thus pointless and not really used
+    # but we set it to this for consistency with other systems.
+    #
+    # Projects installing to sysconfdir, localstatedir or sharedstatedir probably want
+    # to set the following in project():
+    #
+    # default_options : ['sysconfdir=/etc', 'localstatedir=/var', 'sharedstatedir=/var/lib']
         'sysconfdir'        : [ UserStringOption, 'Sysconf data directory.', 'etc' ],
+        'localstatedir'     : [ UserStringOption, 'Localstate data directory.', 'var' ],
+        'sharedstatedir'    : [ UserStringOption, 'Architecture-independent data directory.', 'com' ],
         'werror'            : [ UserBooleanOption, 'Treat warnings as errors.', False ],
         'warning_level'     : [ UserComboOption, 'Compiler warning level to use.', [ '1', '2', '3' ], '1'],
         'layout'            : [ UserComboOption, 'Build directory layout.', ['mirror', 'flat' ], 'mirror' ],
         'default_library'   : [ UserComboOption, 'Default library type.', [ 'shared', 'static' ], 'shared' ],
         'backend'           : [ UserComboOption, 'Backend to use.', backendlist, 'ninja' ],
         'stdsplit'          : [ UserBooleanOption, 'Split stdout and stderr in test logs.', True ],
-        'errorlogs'         : [ UserBooleanOption, "Whether to print the logs from failing tests.", False ],
+        'errorlogs'         : [ UserBooleanOption, "Whether to print the logs from failing tests.", True ],
         }
 
 forbidden_target_names = {'clean': None,
+                          'clean-ctlist': None,
                           'clean-gcno': None,
                           'clean-gcda': None,
                           'coverage-text': None,
@@ -238,10 +247,9 @@ forbidden_target_names = {'clean': None,
                           'all': None,
                           'test': None,
                           'test:': None,
-                          'test-valgrind': None,
-                          'test-valgrind:': None,
                           'benchmark': None,
                           'install': None,
                           'build.ninja': None,
                           'scan-build': None,
+                          'reconfigure': None,
                          }
