@@ -23,7 +23,7 @@ import re
 import os, stat, glob, subprocess, shutil
 import sysconfig
 from collections import OrderedDict
-from . mesonlib import MesonException
+from . mesonlib import MesonException, version_compare, version_compare_many, Popen_safe
 from . import mlog
 from . import mesonlib
 from .environment import detect_cpu_family, for_windows
@@ -65,9 +65,6 @@ class Dependency():
     def need_threads(self):
         return False
 
-    def type_name(self):
-        return self.type_name
-
     def get_pkgconfig_variable(self, variable_name):
         raise MesonException('Tried to get a pkg-config variable from a non-pkgconfig dependency.')
 
@@ -92,7 +89,7 @@ class InternalDependency(Dependency):
         return self.version
 
 class PkgConfigDependency(Dependency):
-    pkgconfig_found = None
+    pkgbin = None
 
     def __init__(self, name, environment, kwargs):
         Dependency.__init__(self, 'pkgconfig')
@@ -109,48 +106,63 @@ class PkgConfigDependency(Dependency):
         else:
             want_cross = environment.is_cross_build()
         self.name = name
-        if PkgConfigDependency.pkgconfig_found is None:
-            self.check_pkgconfig()
+
+        # When finding dependencies for cross-compiling, we don't care about
+        # the 'native' pkg-config
+        if want_cross:
+            if 'pkgconfig' not in environment.cross_info.config['binaries']:
+                if self.required:
+                    raise DependencyException('Pkg-config binary missing from cross file')
+            else:
+                PkgConfigDependency.pkgbin = environment.cross_info.config['binaries']['pkgconfig']
+        # Only search for the native pkg-config the first time and
+        # store the result in the class definition
+        elif PkgConfigDependency.pkgbin is None:
+            PkgConfigDependency.pkgbin = self.check_pkgconfig()
 
         self.is_found = False
-        if not PkgConfigDependency.pkgconfig_found:
+        if not self.pkgbin:
             if self.required:
                 raise DependencyException('Pkg-config not found.')
             return
-        if environment.is_cross_build() and want_cross:
-            if "pkgconfig" not in environment.cross_info.config["binaries"]:
-                raise DependencyException('Pkg-config binary missing from cross file.')
-            pkgbin = environment.cross_info.config["binaries"]['pkgconfig']
+        if want_cross:
             self.type_string = 'Cross'
         else:
-            pkgbin = 'pkg-config'
             self.type_string = 'Native'
 
-        mlog.debug('Determining dependency %s with pkg-config executable %s.' % (name, pkgbin))
-        self.pkgbin = pkgbin
+        mlog.debug('Determining dependency {!r} with pkg-config executable '
+                   '{!r}'.format(name, self.pkgbin))
         ret, self.modversion = self._call_pkgbin(['--modversion', name])
         if ret != 0:
             if self.required:
-                raise DependencyException('%s dependency %s not found.' % (self.type_string, name))
+                raise DependencyException('{} dependency {!r} not found'
+                                          ''.format(self.type_string, name))
             self.modversion = 'none'
             return
-        found_msg = ['%s dependency' % self.type_string, mlog.bold(name), 'found:']
-        self.version_requirement = kwargs.get('version', None)
-        if self.version_requirement is None:
+        found_msg = [self.type_string + ' dependency', mlog.bold(name), 'found:']
+        self.version_reqs = kwargs.get('version', None)
+        if self.version_reqs is None:
             self.is_found = True
         else:
-            if not isinstance(self.version_requirement, str):
-                raise DependencyException('Version argument must be string.')
-            self.is_found = mesonlib.version_compare(self.modversion, self.version_requirement)
+            if not isinstance(self.version_reqs, (str, list)):
+                raise DependencyException('Version argument must be string or list.')
+            if isinstance(self.version_reqs, str):
+                self.version_reqs = [self.version_reqs]
+            (self.is_found, not_found, found) = \
+                version_compare_many(self.modversion, self.version_reqs)
             if not self.is_found:
-                found_msg += [mlog.red('NO'), 'found {!r}'.format(self.modversion),
-                              'but need {!r}'.format(self.version_requirement)]
+                found_msg += [mlog.red('NO'),
+                              'found {!r} but need:'.format(self.modversion),
+                              ', '.join(["'{}'".format(e) for e in not_found])]
+                if found:
+                    found_msg += ['; matched:',
+                                  ', '.join(["'{}'".format(e) for e in found])]
                 if not self.silent:
                     mlog.log(*found_msg)
                 if self.required:
                     raise DependencyException(
                         'Invalid version of a dependency, needed %s %s found %s.' %
-                        (name, self.version_requirement, self.modversion))
+                        (name, not_found, self.modversion))
                 return
         found_msg += [mlog.green('YES'), self.modversion]
         if not self.silent:
@@ -160,18 +172,20 @@ class PkgConfigDependency(Dependency):
         # Fetch the libraries and library paths needed for using this
         self._set_libs()
 
+    def __repr__(self):
+        s = '<{0} {1}: {2} {3}>'
+        return s.format(self.__class__.__name__, self.name, self.is_found,
+                        self.version_reqs)
+
     def _call_pkgbin(self, args):
-        p = subprocess.Popen([self.pkgbin] + args,
-                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
-                             env=os.environ, universal_newlines=True)
-        out = p.communicate()[0]
+        p, out = Popen_safe([self.pkgbin] + args, env=os.environ)[0:2]
         return (p.returncode, out.strip())
 
     def _set_cargs(self):
         ret, out = self._call_pkgbin(['--cflags', self.name])
         if ret != 0:
             raise DependencyException('Could not generate cargs for %s:\n\n%s' % \
-                                      (self.name, out.decode(errors='ignore')))
+                                      (self.name, out))
         self.cargs = out.split()
 
     def _set_libs(self):
@@ -181,7 +195,7 @@ class PkgConfigDependency(Dependency):
         ret, out = self._call_pkgbin(libcmd)
         if ret != 0:
             raise DependencyException('Could not generate libs for %s:\n\n%s' % \
-                                      (self.name, out.decode(errors='ignore')))
+                                      (self.name, out))
         self.libs = []
         for lib in out.split():
             if lib.endswith(".la"):
@@ -223,21 +237,30 @@ class PkgConfigDependency(Dependency):
         return self.libs
 
     def check_pkgconfig(self):
+        evar = 'PKG_CONFIG'
+        if evar in os.environ:
+            pkgbin = os.environ[evar].strip()
+        else:
+            pkgbin = 'pkg-config'
         try:
-            p = subprocess.Popen(['pkg-config', '--version'], stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
-            out = p.communicate()[0]
-            if p.returncode == 0:
-                if not self.silent:
-                    mlog.log('Found pkg-config:', mlog.bold(shutil.which('pkg-config')),
-                             '(%s)' % out.decode().strip())
-                PkgConfigDependency.pkgconfig_found = True
-                return
-        except Exception:
-            pass
-        PkgConfigDependency.pkgconfig_found = False
+            p, out = Popen_safe([pkgbin, '--version'])[0:2]
+            if p.returncode != 0:
+                # Set to False instead of None to signify that we've already
+                # searched for it and not found it
+                pkgbin = False
+        except (FileNotFoundError, PermissionError):
+            pkgbin = False
+        if pkgbin and not os.path.isabs(pkgbin) and shutil.which(pkgbin):
+            # Sometimes shutil.which fails where Popen succeeds, so
+            # only find the abs path if it can be found by shutil.which
+            pkgbin = shutil.which(pkgbin)
         if not self.silent:
-            mlog.log('Found Pkg-config:', mlog.red('NO'))
+            if pkgbin:
+                mlog.log('Found pkg-config:', mlog.bold(pkgbin),
+                         '(%s)' % out.strip())
+            else:
+                mlog.log('Found Pkg-config:', mlog.red('NO'))
+        return pkgbin
 
     def found(self):
         return self.is_found
@@ -289,19 +312,16 @@ class WxDependency(Dependency):
             mlog.log("Neither wx-config-3.0 nor wx-config found; can't detect dependency")
             return
 
-        p = subprocess.Popen([self.wxc, '--version'],
-                             stdout=subprocess.PIPE,
-                             stderr=subprocess.PIPE)
-        out = p.communicate()[0]
+        p, out = Popen_safe([self.wxc, '--version'])[0:2]
         if p.returncode != 0:
             mlog.log('Dependency wxwidgets found:', mlog.red('NO'))
             self.cargs = []
             self.libs = []
         else:
-            self.modversion = out.decode().strip()
+            self.modversion = out.strip()
             version_req = kwargs.get('version', None)
             if version_req is not None:
-                if not mesonlib.version_compare(self.modversion, version_req):
+                if not version_compare(self.modversion, version_req, strict=True):
                     mlog.log('Wxwidgets version %s does not fullfill requirement %s' %\
                              (self.modversion, version_req))
                     return
@@ -310,20 +330,15 @@ class WxDependency(Dependency):
             self.requested_modules = self.get_requested(kwargs)
             # wx-config seems to have a cflags as well but since it requires C++,
             # this should be good, at least for now.
-            p = subprocess.Popen([self.wxc, '--cxxflags'],
-                                 stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
-            out = p.communicate()[0]
+            p, out = Popen_safe([self.wxc, '--cxxflags'])[0:2]
             if p.returncode != 0:
                 raise DependencyException('Could not generate cargs for wxwidgets.')
-            self.cargs = out.decode().split()
+            self.cargs = out.split()
 
-            p = subprocess.Popen([self.wxc, '--libs'] + self.requested_modules,
-                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            out = p.communicate()[0]
+            p, out = Popen_safe([self.wxc, '--libs'] + self.requested_modules)[0:2]
             if p.returncode != 0:
                 raise DependencyException('Could not generate libs for wxwidgets.')
-            self.libs = out.decode().split()
+            self.libs = out.split()
 
     def get_requested(self, kwargs):
         modules = 'modules'
@@ -349,16 +364,14 @@ class WxDependency(Dependency):
     def check_wxconfig(self):
         for wxc in ['wx-config-3.0', 'wx-config']:
             try:
-                p = subprocess.Popen([wxc, '--version'], stdout=subprocess.PIPE,
-                                     stderr=subprocess.PIPE)
-                out = p.communicate()[0]
+                p, out = Popen_safe([wxc, '--version'])[0:2]
                 if p.returncode == 0:
                     mlog.log('Found wx-config:', mlog.bold(shutil.which(wxc)),
-                             '(%s)' % out.decode().strip())
+                             '(%s)' % out.strip())
                     self.wxc = wxc
                     WxDependency.wx_found = True
                     return
-            except Exception:
+            except (FileNotFoundError, PermissionError):
                 pass
         WxDependency.wxconfig_found = False
         mlog.log('Found wx-config:', mlog.red('NO'))
@@ -367,6 +380,8 @@ class WxDependency(Dependency):
         return self.is_found
 
 class ExternalProgram():
+    windows_exts = ('exe', 'com', 'bat')
+
     def __init__(self, name, fullpath=None, silent=False, search_dir=None):
         self.name = name
         if fullpath is not None:
@@ -404,11 +419,10 @@ class ExternalProgram():
             pass
         return False
 
-    @staticmethod
-    def _is_executable(path):
+    def _is_executable(self, path):
         suffix = os.path.splitext(path)[-1].lower()[1:]
         if mesonlib.is_windows():
-            if suffix == 'exe' or suffix == 'com' or suffix == 'bat':
+            if suffix in self.windows_exts:
                 return True
         elif os.access(path, os.X_OK):
             return True
@@ -418,10 +432,15 @@ class ExternalProgram():
         if search_dir is None:
             return False
         trial = os.path.join(search_dir, name)
-        if not os.path.exists(trial):
+        if os.path.exists(trial):
+            if self._is_executable(trial):
+                return [trial]
+        else:
+            for ext in self.windows_exts:
+                trial_ext = '{}.{}'.format(trial, ext)
+                if os.path.exists(trial_ext):
+                    return [trial_ext]
             return False
-        if self._is_executable(trial):
-            return [trial]
         # Now getting desperate. Maybe it is a script file that is a) not chmodded
         # executable or b) we are on windows so they can't be directly executed.
         return self._shebang_to_cmd(trial)
@@ -435,6 +454,11 @@ class ExternalProgram():
         if fullpath or not mesonlib.is_windows():
             # On UNIX-like platforms, the standard PATH search is enough
             return [fullpath]
+        # On Windows, if name is an absolute path, we need the extension too
+        for ext in self.windows_exts:
+            fullpath = '{}.{}'.format(name, ext)
+            if os.path.exists(fullpath):
+                return [fullpath]
         # On Windows, interpreted scripts must have an extension otherwise they
         # cannot be found by a standard PATH search. So we do a custom search
         # where we manually search for a script with a shebang in PATH.
@@ -449,7 +473,7 @@ class ExternalProgram():
         return self.fullpath[0] is not None
 
     def get_command(self):
-        return self.fullpath
+        return self.fullpath[:]
 
     def get_name(self):
         return self.name
@@ -929,10 +953,7 @@ class QtBaseDependency(Dependency):
             if not self.qmake.found():
                 continue
             # Check that the qmake is for qt5
-            pc = subprocess.Popen(self.qmake.fullpath +  ['-v'],
-                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
-                                  universal_newlines=True)
-            stdo = pc.communicate()[0]
+            pc, stdo = Popen_safe(self.qmake.fullpath +  ['-v'])[0:2]
             if pc.returncode != 0:
                 continue
             if not 'Qt version ' + self.qtver in stdo:
@@ -945,9 +966,7 @@ class QtBaseDependency(Dependency):
             return
         self.version = re.search(self.qtver + '(\.\d+)+', stdo).group(0)
         # Query library path, header path, and binary path
-        stdo = subprocess.Popen(self.qmake.fullpath +  ['-query'],
-                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
-                                universal_newlines=True).communicate()[0]
+        stdo = Popen_safe(self.qmake.fullpath +  ['-query'])[1]
         qvars = {}
         for line in stdo.split('\n'):
             line = line.strip()
@@ -1017,8 +1036,9 @@ class QtBaseDependency(Dependency):
         # penalty when using self-built Qt or on platforms
         # where -fPIC is not required. If this is an issue
         # for you, patches are welcome.
-        # Fix this to be more portable, especially to MSVC.
-        return ['-fPIC']
+        if mesonlib.is_linux():
+            return ['-fPIC']
+        return []
 
 class Qt5Dependency(QtBaseDependency):
     def __init__(self, env, kwargs):
@@ -1037,10 +1057,8 @@ class GnuStepDependency(Dependency):
     def detect(self):
         confprog = 'gnustep-config'
         try:
-            gp = subprocess.Popen([confprog, '--help'],
-                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            gp.communicate()
-        except FileNotFoundError:
+            gp = Popen_safe([confprog, '--help'])[0]
+        except (FileNotFoundError, PermissionError):
             self.args = None
             mlog.log('Dependency GnuStep found:', mlog.red('NO'), '(no gnustep-config)')
             return
@@ -1052,20 +1070,12 @@ class GnuStepDependency(Dependency):
             arg = '--gui-libs'
         else:
             arg = '--base-libs'
-        fp = subprocess.Popen([confprog, '--objc-flags'],
-                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        (flagtxt, flagerr) = fp.communicate()
-        flagtxt = flagtxt.decode()
-        flagerr = flagerr.decode()
+        fp, flagtxt, flagerr = Popen_safe([confprog, '--objc-flags'])
         if fp.returncode != 0:
             raise DependencyException('Error getting objc-args: %s %s' % (flagtxt, flagerr))
         args = flagtxt.split()
         self.args = self.filter_arsg(args)
-        fp = subprocess.Popen([confprog, arg],
-                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        (libtxt, liberr) = fp.communicate()
-        libtxt = libtxt.decode()
-        liberr = liberr.decode()
+        fp, libtxt, liberr = Popen_safe([confprog, arg])
         if fp.returncode != 0:
             raise DependencyException('Error getting objc-lib args: %s %s' % (libtxt, liberr))
         self.libs = self.weird_filter(libtxt.split())
@@ -1170,16 +1180,10 @@ class SDL2Dependency(Dependency):
             pass
         sdlconf = shutil.which('sdl2-config')
         if sdlconf:
-            pc = subprocess.Popen(['sdl2-config', '--cflags'],
-                                  stdout=subprocess.PIPE,
-                                  stderr=subprocess.DEVNULL)
-            (stdo, _) = pc.communicate()
-            self.cargs = stdo.decode().strip().split()
-            pc = subprocess.Popen(['sdl2-config', '--libs'],
-                                  stdout=subprocess.PIPE,
-                                  stderr=subprocess.DEVNULL)
-            (stdo, _) = pc.communicate()
-            self.linkargs = stdo.decode().strip().split()
+            pc, stdo = Popen_safe(['sdl2-config', '--cflags'])[0:2]
+            self.cargs = stdo.strip().split()
+            pc, stdo = Popen_safe(['sdl2-config', '--libs'])[0:2]
+            self.linkargs = stdo.strip().split()
             self.is_found = True
             mlog.log('Dependency', mlog.bold('sdl2'), 'found:', mlog.green('YES'), '(%s)' % sdlconf)
             self.version = '2' # FIXME
@@ -1276,18 +1280,7 @@ class Python3Dependency(Dependency):
             pass
         if not self.is_found:
             if mesonlib.is_windows():
-                inc = sysconfig.get_path('include')
-                platinc = sysconfig.get_path('platinclude')
-                self.cargs = ['-I' + inc]
-                if inc != platinc:
-                    self.cargs.append('-I' + platinc)
-                # Nothing exposes this directly that I coulf find
-                basedir = sysconfig.get_config_var('base')
-                vernum = sysconfig.get_config_var('py_version_nodot')
-                self.libs = ['-L{}/libs'.format(basedir),
-                             '-lpython{}'.format(vernum)]
-                self.is_found = True
-                self.version = sysconfig.get_config_var('py_version_short')
+                self._find_libpy3_windows(environment)
             elif mesonlib.is_osx():
                 # In OSX the Python 3 framework does not have a version
                 # number in its name.
@@ -1301,6 +1294,42 @@ class Python3Dependency(Dependency):
         else:
             mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.red('NO'))
 
+    def _find_libpy3_windows(self, env):
+        '''
+        Find python3 libraries on Windows and also verify that the arch matches
+        what we are building for.
+        '''
+        pyarch = sysconfig.get_platform()
+        arch = detect_cpu_family(env.coredata.compilers)
+        if arch == 'x86':
+            arch = '32'
+        elif arch == 'x86_64':
+            arch = '64'
+        else:
+            # We can't cross-compile Python 3 dependencies on Windows yet
+            mlog.log('Unknown architecture {!r} for'.format(arch),
+                     mlog.bold(self.name))
+            self.is_found = False
+            return
+        # Pyarch ends in '32' or '64'
+        if arch != pyarch[-2:]:
+            mlog.log('Need', mlog.bold(self.name),
+                     'for {}-bit, but found {}-bit'.format(arch, pyarch[-2:]))
+            self.is_found = False
+            return
+        inc = sysconfig.get_path('include')
+        platinc = sysconfig.get_path('platinclude')
+        self.cargs = ['-I' + inc]
+        if inc != platinc:
+            self.cargs.append('-I' + platinc)
+        # Nothing exposes this directly that I coulf find
+        basedir = sysconfig.get_config_var('base')
+        vernum = sysconfig.get_config_var('py_version_nodot')
+        self.libs = ['-L{}/libs'.format(basedir),
+                     '-lpython{}'.format(vernum)]
+        self.version = sysconfig.get_config_var('py_version_short')
+        self.is_found = True
+
     def get_compile_args(self):
         return self.cargs
 
